(* $Id: Path.v,v 1.7 2012/02/10 09:11:50 garrigue Exp $ *)

Set Implicit Arguments.
Require Import AdditionalTactics LibTactics List Arith.

(* A simple tactic for case analysis *)
Ltac case_rewrite H t :=
  case_eq t; introv H; rewrite H in *; try discriminate.

(* Lists *)

Inductive list_forall (A : Set) (P : A -> Prop) : list A -> Prop :=
  | list_forall_nil : list_forall P nil
  | list_forall_cons : forall L x,
       list_forall P L -> P x -> list_forall P (x::L).
Hint Constructors list_forall.

(* Domain of an association list *)

Section Dom.
Variables A B : Set.
Fixpoint dom (l:list(A*B)) :=
  match l with
  | nil => nil
  | (a,b) :: l' => a :: dom l'
  end.

Lemma in_dom : forall a b l,
  In (a,b) l -> In a (dom l).
Proof.
  induction l; simpl; intros.
  elim H.
  destruct H; subst; simpls*.
  destruct a0; simpls*.
Qed.
End Dom.

(* Start of definitions for the path calculus *)

(* Definition of access signature and path *)

Definition var := nat.
Inductive name : Set := Name : nat -> name.

Inductive signature : Set :=
  | Snil  : signature
  | Scons : name -> signature -> signature -> signature.

Definition env := list signature.

Inductive In_sig : list name -> signature -> Prop :=
  | InS_nil : forall S, In_sig nil S
  | InS_hd : forall m1 ap S1 S,
      In_sig ap S1 -> In_sig (m1::ap) (Scons m1 S1 S)
  | InS_tl : forall m1 ap S1 S,
      In_sig ap S -> In_sig ap (Scons m1 S1 S).
Hint Constructors In_sig.

Inductive path : Set :=
  | Prt  : path
  | Pvar : var -> signature -> path
  | Pdot : path -> name -> path
  | Papp : path -> path -> path.

(* Substitution and properties *)

Definition subs := list path.

Fixpoint subst (s:subs) (p:path) {struct p} : path :=
  match p with
  | Prt => Prt
  | Pvar v sg => nth v s p
  | Pdot p1 nm => Pdot (subst s p1) nm
  | Papp p1 p2 => Papp (subst s p1) (subst s p2)
  end.

Fixpoint app_path (p:path) (ap:list name) {struct ap} : path :=
  match ap with
  | nil => p
  | m::ap' => app_path (Pdot p m) ap'
  end.

Lemma app_path_app : forall ap2 ap1 p,
  app_path p (ap1 ++ ap2) = app_path (app_path p ap1) ap2.
Proof.
  induction ap1; intros; simpls*.
Qed.

Lemma app_path_eq : forall n sg n' sg' ap ap',
  app_path (Pvar n sg) ap = app_path (Pvar n' sg') ap' ->
  n = n' /\ sg = sg' /\ ap = ap'.
Proof.
  induction ap using rev_ind; induction ap' using rev_ind;
    rewrite_all app_path_app; simpl; intros; try discriminate.
    inverts* H.
  inverts H.
  destruct (IHap _ H1) as [Eq1 [Eq2 Eq3]].
  rewrite* Eq3.
Qed.

Lemma subst_vp : forall s n sg ap,
  subst s (app_path (Pvar n sg) ap) = app_path (subst s (Pvar n sg)) ap.
Proof.
  induction ap using rev_ind. auto.
  rewrite_all app_path_app. simpl. rewrite* IHap.
Qed.

(* Well-formed finite map *)

Section Ok.
Variable A : Set.

Inductive ok : list (name*A) -> Prop :=
  | ok_nil  : ok nil
  | ok_cons : forall m a l,
      ok l -> ~In m (dom l) -> ok ((m,a)::l).

Lemma ok_in_func : forall m a b l,
  ok l -> In (m,a) l -> In (m,b) l -> a = b.
Proof.
  induction 1; intros. elim H.
  simpl in *.
  destruct H1.
  Case "hd".
    inverts H1.
    destruct H2.
    SCase "hd". inverts~ H1.
    SCase "tl". elim (H0 (in_dom _ _ _ H1)).
  Case "tl".
    destruct H2.
    SCase "hd". inverts H2. elim (H0 (in_dom _ _ _ H1)).
    SCase "tl". auto.
Qed.
End Ok.

(* Definition of well-formed programs *)

Inductive expr : Set :=
  | Fun : signature -> expr -> expr
  | Mod : list (name * expr) -> expr
  | Abb : path -> expr.

Inductive is_nth (A:Set) : nat -> A -> list A -> Prop :=
  | is_nthO : forall a l, is_nth 0 a (a::l)
  | is_nthS : forall n b a l,
      is_nth n a l -> is_nth (S n) a (b::l).

Inductive closed (e:env) : path -> Prop :=
  | closed_rt  : closed e Prt
  | closed_var : forall n sg, is_nth n sg e -> closed e (Pvar n sg)
  | closed_dot : forall p m, closed e p -> closed e (Pdot p m)
  | closed_app : forall p1 p2,
      closed e p1 -> closed e p2 -> closed e (Papp p1 p2).
Hint Constructors closed.

Lemma closed_app_inv : forall e ap p,
  closed e (app_path p ap) -> closed e p.
Proof.
  induction ap using rev_ind; intros. auto.
  rewrite app_path_app in H. inverts* H.
Qed.

Inductive program (e:env) : expr -> Prop :=
  | prog_fun : forall sg d, program (sg::e) d -> program e (Fun sg d)
  | prog_mod : forall l,
      ok l ->
      (forall m d, In (m,d) l -> program e d) ->
      program e (Mod l)
  | prog_abb : forall p, closed e p -> program e (Abb p).

Inductive nonpath : expr -> Prop :=
  | np_fun : forall sg e, nonpath (Fun sg e)
  | np_mod : forall l, nonpath (Mod l).
Hint Constructors nonpath.

Fixpoint rooted_path (p : path) : bool :=
  match p with
  | Prt => true
  | Pvar _ _ => false
  | Pdot p _ => rooted_path p
  | Papp p1 p2 => rooted_path p1
  end.

Section Program.

(* Fix a well-formed program define safe reduction *)

Variable P : expr.
Hypothesis P_ok : program nil P.

(* Lookup and properties *)

Inductive lookup : path -> env -> subs -> expr -> Prop :=
  | lookup_rt  : lookup Prt nil nil P
  | lookup_dot : forall p e m s l d,
      lookup p e s (Mod l) -> In (m,d) l -> lookup (Pdot p m) e s d
  | lookup_app : forall p1 e p2 s sg d,
      lookup p1 e s (Fun sg d) -> lookup (Papp p1 p2) (sg::e) (p2::s) d.
Hint Constructors lookup.

Lemma lookup_program : forall p e s d,
  lookup p e s d -> program e d.
Proof.
  induction 1; intros.
  Case "rt". auto.
  Case "dot". inverts* IHlookup.
  Case "app". inverts* IHlookup.
Qed.

Lemma lookup_closed : forall p e s q,
  lookup p e s (Abb q) -> closed e q.
Proof.
  intros.
  lets Prog: lookup_program H.
  inverts* Prog.
Qed.
Hint Immediate lookup_closed.

Lemma lookup_args_closed : forall p e s d e0,
  lookup p e s d -> closed e0 p -> list_forall (closed e0) s.
Proof.
  induction 1; intros. auto.
  inverts* H1.
  inverts* H0.
Qed.
Hint Immediate lookup_args_closed.

Lemma lookup_length : forall p e s d,
  lookup p e s d -> length e = length s.
Proof.
  induction 1; simpls*.
Qed.
Hint Resolve lookup_length.

Lemma lookup_subst : forall s p e s0 d,
  lookup p e s0 d -> lookup (subst s p) e (map (subst s) s0) d.
Proof.
  induction 1; simpls*.
Qed.
Hint Resolve lookup_subst.

Lemma lookup_unique : forall p e s d e' s' d',
  lookup p e s d -> lookup p e' s' d' -> e = e' /\ s = s' /\ d = d'.
Proof.
  introv Lk.
  revert e' s' d'; induction Lk; intros.
  Case "rt".
  inverts~ H.
  Case "dot".
  inverts H0.
  forwards Eq:(IHLk _ _ _ H3).
  destruct Eq as [E1 [E2 E3]].
  inverts E3. subst e' s'.
  splits~.
  lets l1_ok: lookup_program Lk.
  inverts l1_ok.
  apply* ok_in_func.
  Case "app".
  inverts H.
  forwards Eq:(IHLk _ _ _ H5).
  destruct Eq as [E1 [E2 E3]].
  inverts E3. subst e0 s0.
  splits~.
Qed.

Lemma lookup_rooted : forall p e s d,
  lookup p e s d -> rooted_path p = true.
Proof.
  induction 1; simpls*.
Qed.

Lemma rooted_vp : forall n sg ap,
  rooted_path (app_path (Pvar n sg) ap) = false.
Proof.
  induction ap using rev_ind; simpls*.
  rewrite* app_path_app.
Qed.

(* Safe reduction, safety wrt a signature, safe substitution *)

Inductive red : path -> path -> Prop :=
  | r_src : forall p e s d,
      lookup p e s d -> safe_subs s e -> nonpath d -> red p p
  | r_vp : forall n sg ap,
      In_sig ap sg -> red (app_path (Pvar n sg) ap) (app_path (Pvar n sg) ap)
  | r_exp : forall p e s p' q,
      lookup p e s (Abb p') -> safe_subs s e -> red (subst s p') q -> red p q
  | r_dot : forall p p' m q,
      red p p' -> red (Pdot p' m) q -> red (Pdot p m) q
  | r_app : forall p1 p1' p2 q,
      red p1 p1' -> red (Papp p1' p2) q -> red (Papp p1 p2) q
with safe : path -> signature -> Prop :=
  | safe_nil : forall p q,
      red p q -> safe p Snil
  | safe_cons : forall p m1 S1 S,
      safe p S -> safe (Pdot p m1) S1 -> safe p (Scons m1 S1 S)
with safe_subs : subs -> env -> Prop :=
  | sfsubs_nil : safe_subs nil nil
  | sfsubs_cons : forall sg e p s,
      safe p sg -> safe_subs s e -> safe_subs (p::s) (sg::e).

Hint Constructors red safe safe_subs.

(* Define mutual induction principles *)

Scheme red_ind3 := Minimality for red Sort Prop
  with safe_ind3 := Minimality for safe Sort Prop
  with safe_subs_ind3 := Minimality for safe_subs Sort Prop.

(* Properties of safety and reduction *)

Lemma safe_red : forall p S,
  safe p S -> exists q, red p q.
Proof.
  induction 1; intros. exists* q.
  auto.
Qed.

Lemma safe_all : forall p S,
  (forall ap, In_sig ap S -> exists q, red (app_path p ap) q) -> safe p S.
Proof.
  intros.
  gen p; induction S; intros.
  Case "S = Snil".
    destruct* (H nil).
  Case "S <> Snil".
    constructor.
      apply IHS2; intros. destruct* (H ap).
    forwards Safe: IHS1 (Pdot p n).
      intros; destruct* (H (n::ap)).
    auto.
Qed.

Lemma safe_inv : forall p S ap,
  safe p S -> In_sig ap S -> exists q, red (app_path p ap) q.
Proof.
  intros.
  gen p; induction H0; simpl; intros.
  Case "nil". apply* safe_red.
  Case "hd". apply* IHIn_sig. inverts* H.
  Case "tl". apply* IHIn_sig. inverts* H.
Qed.

Lemma red_unique_src : forall p q' e s d,
  red p q' -> lookup p e s d -> nonpath d -> p = q'.
Proof.
  introv Red.
  revert e s d; induction Red; intros; auto.
  Case "r_exp".
    destruct (lookup_unique H1 H) as [E1 [E2 E3]]. subst; clear H1.
    inversion H2.
  Case "r_dot".
    inverts keep H.
    rewrite* <- (IHRed1 _ _ _ H3) in IHRed2.
  Case "r_app".
    inverts keep H.
    rewrite* <- (IHRed1 _ _ _ H6) in IHRed2.
Qed.

Lemma red_idem : forall p q,
  red p q -> red q q.
Proof.
  induction 1; auto*.
Qed.

Lemma lookup_vp_false : forall n sg ap e s d,
  ~lookup (app_path (Pvar n sg) ap) e s d.
Proof.
  induction ap using rev_ind; simpl; introv Lk; inverts Lk;
    rewrite_all app_path_app in *; simpl in *; try discriminate.
  inverts* H.
Qed.

(* Uniqueness of reduction *)

Lemma red_unique_vp : forall n sg ap q',
  red (app_path (Pvar n sg) ap) q' ->
  In_sig ap sg /\ app_path (Pvar n sg) ap = q'.
Proof.
  intros.
  sets_eq p:(app_path (Pvar n sg) ap).
  gen ap; induction H; intros; subst*.
  Case "r_src". elim (lookup_vp_false _ _ _ H).
  Case "r_vp".
    destruct (app_path_eq _ _ _ _ _ _ EQp) as [Eq1 [Eq2 Eq3]].
    subst*.
  Case "r_exp". elim (lookup_vp_false _ _ _ H).
  Case "r_dot".
    assert (p = p').
      induction ap using rev_ind. discriminate.
      clear IHap.
      rewrite app_path_app in *; simpl in *.
      inverts EQp.
      destruct* (IHred1 ap).
    subst.
    apply* IHred2.
  Case "r_app".
    induction ap using rev_ind. discriminate.
    rewrite app_path_app in *; discriminate.
Qed.

Lemma red_unique : forall p q q',
  red p q -> red p q' -> q = q'.
Proof.
  intros until 1.
  revert q'.
  induction H; intros.
  Case "r_src". apply* red_unique_src.
  Case "r_vp". destruct* (red_unique_vp _ _ _ H0).
  Case "r_exp".
    induction H2; auto.
    SCase "r_src". rewrite* (red_unique_src (r_exp H H0 H1) H2 H4).
    SCase "r_vp". elim (lookup_vp_false _ _ _ H).
    SCase "r_exp".
      destruct (lookup_unique H2 H) as [E1 [E2 E3]]; subst.
      inverts* E3.
    SCase "r_dot".
      inverts keep H.
      rewrite <- (red_unique_src H2_ H4) in *; auto.
    SCase "r_app".
      inverts keep H.
      rewrite <- (red_unique_src H2_ H7) in *; auto.
  Case "r_dot".
    apply* IHred2.
    clear IHred2.
    inverts H1.
    SCase "r_src".
      inverts keep H2.
      rewrite* <- (red_unique_src H H6).
    SCase "r_vp".
      lets Red: (r_dot H H0). rewrite <- H2 in Red.
      rewrite* (proj2 (red_unique_vp _ _ _ Red)).
    SCase "r_exp".
      inverts keep H2.
      rewrite* <- (red_unique_src H H6).
    SCase "r_dot".
      rewrite* (IHred1 _ H4).
  Case "r_app".
    apply* IHred2; clear IHred2.
    inverts H1.
    SCase "r_src".
      inverts keep H2.
      rewrite* <- (red_unique_src H H9).
    SCase "r_vp".
      induction ap using rev_ind. discriminate.
      rewrite app_path_app in H2. discriminate.
    SCase "r_exp".
      inverts keep H2.
      rewrite* <- (red_unique_src H H9).
    SCase "r_app".
      rewrite* (IHred1 _ H4).
Qed.

(* Equivalence of converging paths *)

Lemma safe_equiv : forall p p' q S,
  safe p S -> red p q -> red p' q -> safe p' S.
Proof.
  introv Safe; gen p' q; induction Safe; intros.
  Case "Snil". auto*.
  Case "Scons".
    constructor; auto*.
    destruct (safe_red Safe2).
    apply* IHSafe2.
    inverts keep H1.
    SCase "r_src".
      inverts keep H2.
      rewrite* <- (red_unique_src H H7) in H0.
    SCase "r_vp".
      apply (r_dot H0).
      induction ap using rev_ind. discriminate.
      clear IHap.
      rewrite app_path_app in *; simpl in *.
      inverts H2.
      rewrite* <- (proj2 (red_unique_vp _ _ _ H)).
    SCase "r_exp".
      inverts keep H2.
      rewrite* <- (red_unique_src H H7) in H0.
    SCase "r_dot".
      rewrite* (red_unique H4 H) in H6.
Qed.

(* Variable paths *)

Lemma vp_path_red : forall p q,
  red p q -> rooted_path p = false -> p = q.
Proof.
  induction 1; simpl; intros; auto*.
  Case "r_exp". rewrite (lookup_rooted H) in H2. discriminate.
  Case "r_dot". rewrite <- IHred1 in *; auto.
  Case "r_app". rewrite <- IHred1 in *; auto.
Qed.

Lemma root_path_red : forall p q,
  red p q -> rooted_path p = false ->
  exists n, exists sg, exists ap, p = app_path (Pvar n sg) ap.
Proof.
  intros. induction* H.
  Case "r_src".
    rewrite* (lookup_rooted H) in H0; discriminate.
  Case "r_exp".
    rewrite* (lookup_rooted H) in H0; discriminate.
  Case "r_dot".
    destruct* IHred1 as [n [sg [ap Eq]]].
    exists n; exists sg; exists (ap++m::nil).
    rewrite Eq. rewrite* app_path_app.
  Case "r_app".
    simpl in *.
    destruct* IHred1 as [n [sg [ap Eq]]].
    subst.
    rewrite (proj2 (red_unique_vp _ _ _ H)) in *.
    auto.
Qed.

Lemma safe_subs_nth : forall d e s n sg,
  closed e (Pvar n sg) -> safe_subs s e -> safe (nth n s d) sg.
Proof.
  intros.
  inverts H.
  gen s; induction H2; intros; inverts H0; simpls*.
Qed.

(* Substitution on closed paths *)

Lemma subst_map : forall s s0 e p,
  closed e p -> length e = length s0 ->
  subst (map (subst s) s0) p = subst s (subst s0 p).
Proof.
  induction 1; simpls*; intros.
  Case "var".
    set (d:= Pvar n sg). clearbody d.
    gen s0; induction H; intros; destruct s0; try discriminate; simpls*.
  Case "dot". rewrite* IHclosed.
  Case "app". rewrite* IHclosed1. rewrite* IHclosed2.
Qed.

Lemma subst_closed : forall s e p e',
  list_forall (closed e') s -> closed e p -> length e = length s ->
  closed e' (subst s p).
Proof.
  induction 2; simpls*.
  set (d:=Pvar n sg). clearbody d.
  gen s; induction H0; intros; inverts H; simpls*; discriminate.
Qed.
Hint Resolve subst_closed.

Lemma red_closed : forall e p q,
  closed e p -> red p q -> closed e q.
Proof.
  induction 2; auto*.
  Case "r_dot". inverts* H.
  Case "r_app". inverts* H.
Qed.
Hint Immediate red_closed.

(* The substitution lemma *)

Lemma substitution : forall p S s e,
  safe p S -> safe_subs s e -> closed e p ->
  safe (subst s p) S /\
  forall q, red p q -> if rooted_path q then red (subst s p) (subst s q)
                       else exists r, red (subst s p) r /\ red (subst s q) r.
Proof.
  introv Safe Subs Cls.
  revert p S Safe Cls.
  refine (safe_ind3
    (fun p1 p2 => closed e p1 ->
                  if rooted_path p2 then red (subst s p1) (subst s p2)
                  else exists q, red (subst s p1) q /\ red (subst s p2) q) _
    (fun s' e' => list_forall (closed e) s' ->
                  safe_subs (map (subst s) s') e') _ _ _ _ _ _ _ _ _);
    intros.
  Case "r_src".
    case_eq (rooted_path p); introv R1.
    assert (lookup (subst s p) e0 (map (subst s) s0) d).
      clear -H; induction H; simpls*.
    apply* r_src.
    rewrite (lookup_rooted H) in R1; discriminate.
  Case "r_vp".
      rewrite rooted_vp.
      intros.
      rewrite subst_vp. simpl.
      destruct* (@safe_inv (nth n s (Pvar n sg)) sg ap).
      apply* safe_subs_nth.
      apply* closed_app_inv.
  Case "r_exp".
    lets Cls: lookup_closed H.
    lets Cls': lookup_args_closed H H4.
    case_rewrite R1 (rooted_path q); intros.
    SCase "true".
      apply* r_exp. auto*.
      rewrite* (subst_map s _ Cls (lookup_length H)).
    SCase "false".
      lets Red: (r_exp H H0 H2).
      lets Lk: lookup_subst s H.
      rewrite <- (subst_map s _ Cls (lookup_length H)) in H3 by auto.
      destruct* H3 as [q0 [R2 R3]].
  Case "r_dot".
    inverts H3.
    case_rewrite R1 (rooted_path q).
    SCase "true".
      simpl in *.
      case_rewrite R2 (rooted_path p').
      SSCase "true". apply* r_dot.
      SSCase "false".
        rewrite* <- (vp_path_red H1) in R1.
        simpl in R1. rewrite R1 in R2; discriminate.
    SCase "false".
      destruct* H2 as [q0 [R3 R4]].
      exists q0.
      split*.
      simpl in *.
      case_rewrite R2 (rooted_path p').
      SSCase "true".
        apply* r_dot.
      SSCase "false".
        destruct* H0 as [q1 [R5 R6]].
        assert (safe q1 (Scons m Snil Snil)).
          refine (safe_equiv _ R6 _).
            constructor; apply* safe_nil.
          apply* red_idem.
        inverts H0.
        inverts H8.
        lets Red: r_dot R5 H0.
        lets Red': r_dot R6 H0.
        rewrite (red_unique R3 Red').
        assumption.
  Case "r_app".
    case_rewrite R2 (rooted_path p1').
    SCase "true".
      simpl in *.
      inverts H3.
      lets Cls: red_closed H6 H.
      case_rewrite R1 (rooted_path q).
      SSCase "true".
        apply* r_app.
      SSCase "false".
        destruct* H2 as [q0 [R3 R4]].
    SCase "false".
      destruct* (root_path_red H1) as [n [sg [ap Eq]]].
      induction ap using rev_ind. discriminate.
      rewrite app_path_app in Eq. discriminate.
  Case "safe_nil".
    case_rewrite R1 (rooted_path q); intuition eauto;
      try rewrite (red_unique H0 H) in *; try rewrite R1.
        assumption.
      destruct H1 as [q0 [Rp Rq]].
      apply* safe_nil.
    assumption.
  Case "safe_cons".
    destruct (H0 Cls).
    split*.
    constructor; auto.
    forwards Safe: H2; auto*.
  Case "sfsubs_nil". simpls*.
  Case "sfsubs_cons". simpl. inverts* H3.
Qed.

End Program.

Check substitution.
