(* $Id: Path2.v,v 1.14 2012/02/09 04:48:07 garrigue Exp $ *)

Set Implicit Arguments.
Require Import AdditionalTactics LibTactics List Arith Omega Path.

(* Error monad and association lists *)

Section Bind.
Variable A B : Type.
Definition bind (x : option A) (f : A -> option B) :=
  match x with
  | None => None
  | Some y => f y
  end.
End Bind.

Ltac case_bind v :=
  match goal with
    H: context [bind ?exp ?cont] |- _ =>
      let Eq := fresh "Eq" in
      case_eq exp; [intros v Eq | intros Eq];
      rewrite Eq in *; try discriminate; simpl in H
  end.

Section Assoc.
Variable A B : Set.
Variable beq : A -> A -> bool.
Hypothesis eq_dec : forall x y, beq x y = true <-> x = y.

Fixpoint assoc x (l : list (A * B)) :=
  match l with
  | nil => None
  | (a,b)::rem => if beq x a then Some b else assoc x rem
  end.

Lemma assoc_sound : forall x y l,
  assoc x l = Some y -> In (x,y) l.
Proof.
  induction l; simpl; intros. discriminate.
  destruct a.
  revert H; case_eq (beq x a); intros; auto.
  assert (x = a) by apply* eq_dec.
  inversions* H0.
Qed.

Lemma assoc_complete : forall x y l,
  assoc x l = None -> ~In (x,y) l.
Proof.
  induction l; simpl; intros; intro Hin. tauto.
  destruct a.
  destruct Hin.
    inverts H0.
    replace (beq x x) with true in H. discriminate.
    symmetry; apply* eq_dec.
  destruct (beq x a). discriminate. intuition.
Qed.

Lemma assoc_app : forall x l1 l2,
  assoc x (l1 ++ l2) =
  match assoc x l1 with
  | None => assoc x l2
  | Some b => Some b
  end.
Proof.
  intros; induction l1; simpl; auto.
  destruct a.
  destruct* (beq x a).
Qed.

Lemma assoc_app2 : forall x l1 l2,
  assoc x l1 = None -> assoc x (l1 ++ l2) = assoc x l2.
Proof.
  intros.
  rewrite assoc_app.
  rewrite* H.
Qed.

Variable f : B -> B.

Definition map_snd := map (fun (x:A*B) => let (a,b) := x in (a,f b)).

Lemma map_assoc : forall a l,
  assoc a (map_snd l) = bind (assoc a l) (fun b => Some (f b)).
Proof.
  induction l; simpl. auto.
  destruct a0.
  destruct* (beq a a0).
Qed.
End Assoc.

Hint Resolve assoc_sound.

(* Properties of is_nth *)

Section Nth.
Variable A : Set.

Lemma is_nth_lt : forall v (s:A) e, is_nth v s e -> v < length e.
Proof.
  induction 1; simpl; auto with arith.
Qed.

Lemma is_nth_nth : forall v (s:A) e (d:A), is_nth v s e -> nth v e d = s.
Proof.
  induction 1; simpl; auto with arith.
Qed.

Lemma nth_is_nth : forall v e (d:A),
  v < length e -> is_nth v (nth v e d) e.
Proof.
  intros; gen v; induction e; simpl; intros.
    elimtype False; omega.
  destruct v; constructor.
  apply IHe. omega.
Qed.
End Nth.

(* Equality on names, paths an locations *)

Scheme Equality for name.

Lemma name_dec_blq : forall x y, name_beq x y = true <-> x = y.
Proof.
  destruct x; destruct y.
  simpl. split; intros.
    f_equal. apply* nat_dec_bl.
  inversion H. apply* nat_dec_lb.
Qed.
Hint Resolve name_dec_blq.

Lemma name_beq_refl : forall x, name_beq x x = true.
Proof.
  induction x.
  simpl.
  apply* nat_dec_lb.
Qed.

Lemma assoc_ok_complete : forall (B:Set) x (y:B) l,
  ok l -> In (x,y) l -> assoc name_beq x l = Some y.
Proof.
  induction 1; simpl; intros.
    elim H.
  destruct H1.
    inverts H1.
    rewrite* name_beq_refl.
  case_eq (name_beq x m); intros.
    elim H0.
    rewrite <- (proj1 (name_dec_blq _ _) H2).
    apply* in_dom.
  auto.
Qed.

Fixpoint path_beq (p1 p2 : path) :=
  match (p1, p2) with
  | (Prt, Prt) => true
  | (Pvar x _, Pvar y _) => nat_beq x y
  | (Pdot q1 m, Pdot q2 n) =>
    if name_eq_dec m n then path_beq q1 q2 else false
  | (Papp q1 q1', Papp q2 q2') =>
    if path_beq q1 q2 then path_beq q1' q2' else false
  | _ => false
  end.

Lemma path_beq_refl : forall p, path_beq p p = true.
Proof.
  induction p; simpl; auto.
  Case "Pvar".
    rewrite nat_dec_lb; auto.
  Case "Pdot".
    rewrite IHp.
    destruct (name_eq_dec n n); auto.
  Case "Papp".
    rewrite IHp1, IHp2. auto.
Qed.

Lemma nat_beq_sym : forall m n, nat_beq m n = nat_beq n m.
Proof.
  induction m; destruct n; simpl; auto.
Qed.

Lemma name_beq_sym : forall m n, name_beq m n = name_beq n m.
Proof.
  induction m as [v]; destruct n; simpl; intros; auto using nat_beq_sym.
Qed.

Lemma path_beq_sym : forall p q, path_beq p q = path_beq q p.
Proof.
  induction p; destruct q; simpl; auto using nat_beq_sym.
    rewrite IHp.
    destruct (name_eq_dec n n0); destruct (name_eq_dec n0 n);
      auto; elimtype False; intuition.
  rewrite IHp1, IHp2.
  auto.
Qed.

Scheme Equality for list.

Definition list_name_eq_dec : forall x y : list name, {x=y}+{x<>y}.
  intros.
  apply (list_eq_dec name_beq); intros; apply* name_dec_blq.
Defined.

(* Properties of paths *)

Fixpoint path_length (p : path) :=
  match p with
  | Prt | Pvar _ _ => 0
  | Pdot p1 _ => S (path_length p1)
  | Papp p1 _ => S (path_length p1)
  end.

Lemma app_path_length : forall p ap,
  path_length (app_path p ap) = path_length p + length ap.
Proof.
  intros; revert p.
  induction ap; simpl; intros.
    auto with arith.
  rewrite IHap.
  simpl.
  auto with arith.
Qed.

Lemma path_beq_length : forall p q,
  path_beq p q = true -> path_length p = path_length q.
Proof.
  induction p; destruct q; simpl; intros; try discriminate; auto.
    destruct (name_eq_dec n n0).
      rewrite (IHp _ H). auto.
    discriminate.
  specialize (IHp1 q1).
  destruct (path_beq p1 q1).
    rewrite IHp1; auto.
  discriminate.
Qed.

Section Program.

(* Fix a well-formed program, and define semi-ground normalization *)

Variable P : expr.
Hypothesis P_ok : program nil P.

(* Algorithmic version of lookup *)

Inductive lookup_res :=
  | Lk_ok : env -> subs -> expr -> list name -> lookup_res
  | Lk_var : var -> signature -> list name -> lookup_res
  | Lk_fail : lookup_res.

Fixpoint lookup_dec (q : path) :=
  match q with
  | Prt => Lk_ok nil nil P nil
  | Pvar x sg => Lk_var x sg nil
  | Pdot q1 m =>
    match lookup_dec q1 with
    | Lk_ok e s d loc =>
      match d with
      | Mod l =>
        match assoc name_beq m l with
        | None => Lk_fail
        | Some d' => Lk_ok e s d' (app loc (m :: nil))
        end
      | _ => Lk_fail
      end
    | Lk_var x sg ap => Lk_var x sg (app ap (m :: nil))
    | Lk_fail => Lk_fail
    end
  | Papp q1 q2 =>
    match lookup_dec q1 with
    | Lk_ok e s d loc =>
      match d with
      | Fun sg d' => Lk_ok (sg::e) (q2::s) d' loc
      | _ => Lk_fail
      end
    | _ => Lk_fail
    end
  end.

Lemma lookup_sound : forall p e s d loc,
  lookup_dec p = Lk_ok e s d loc -> lookup P p e s d.
Proof.
  induction p; simpl; intros.
  Case "Prt".
    inversion H; constructor.
  Case "Pvar".
    inversion H.
  Case "Pdot".
    revert H.
    case_eq (lookup_dec p); intros; try solve [inversion H0].
    destruct e1; try solve [inversion H0].
    SCase "Mod".
    revert H0; case_eq (assoc name_beq n l0); intros; inversions H1.
    apply lookup_dot with (l:=l0); eauto.
  Case "Papp".
    revert H.
    case_eq (lookup_dec p1); intros; try solve [inversion H0].
    destruct e1; inversions H0.
    SCase "Fun".
    constructor. eauto.
Qed.
Hint Resolve lookup_sound lookup_length.

Lemma lookup_complete : forall p e s d,
  lookup P p e s d -> exists ap, lookup_dec p = Lk_ok e s d ap.
Proof.
  induction 1; simpl; intros.
  Case "rt".
    esplit; auto*.
  Case "dot".
    destruct IHlookup as [ap lkp].
    rewrite lkp.
    lets Prog: (lookup_program P_ok H).
    inverts Prog.
    rewrite (assoc_ok_complete _ _ H2 H0).
    esplit; auto*.
  Case "app".
    destruct IHlookup as [ap lkp].
    rewrite lkp.
    esplit; auto.
Qed.

(* Decide membership in an access signature *)

Lemma In_sig_dec : forall ap sg, {In_sig ap sg}+{~In_sig ap sg}.
Proof.
  intros; revert ap.
  induction sg; intros; destruct ap; try solve [left; constructor].
  Case "Snil".
    right; intro.
    inversion H.
  Case "SCons".
    destruct* (IHsg2 (n0::ap)).
    destruct (name_eq_dec n0 n).
    SCase "n0=n".
      subst.
      destruct* (IHsg1 ap).
      right.
      intro Hsg; inversion* Hsg.
    SCase "n0<>n".
      right.
      intro Hsg; inversion* Hsg.
Qed.

Section Sgnlz.

(* Define semi-ground normalization by mutual recursion,
   closing the knot on expand *)

Variable expand : path -> option path.

Fixpoint sgnlz (q : path) :=
  match q with
  | Prt | Pvar _ _ => expand q
  | Pdot q1 m  => bind (sgnlz q1) (fun q1' => expand (Pdot q1' m))
  | Papp q1 q2 => bind (sgnlz q1) (fun q1' => expand (Papp q1' q2))
  end.

Definition head_subst (T:subs) (R:list (path * path)) (p:path) :=
  if rooted_path p then subst T p else
  match assoc path_beq p R with None => p | Some q => q end.

Fixpoint sig_subs vp p (S : signature) :=
  match S with
  | Snil => bind (sgnlz p) (fun q => Some ((vp, q) :: nil))
  | Scons m S1 S' =>
    bind (sig_subs (Pdot vp m) (Pdot p m) S1)
    (fun l => bind (sig_subs vp p S') (fun l' => Some (app l l')))
  end.

Fixpoint vp_subs n (e:env) (s:subs) :=
  match (e,s) with 
  | (sg :: e', p :: s') =>
    bind (sig_subs (Pvar n sg) p sg)
    (fun l => bind (vp_subs (S n) e' s') (fun l' => Some (app l l')))
  | (nil, nil) => Some nil
  | _ => None
  end.

End Sgnlz.

Fixpoint expand (h : nat) (lck : list (list name)) p :=
  match h with
  | 0 => None
  | S h' =>
    match lookup_dec p with
    | Lk_var x sg ap =>
      if In_sig_dec ap sg then Some p else None
    | Lk_ok e s d loc =>
      bind (vp_subs (expand h' lck) 0 e s)
      (fun R =>
        match d with
        | Abb q =>
          if In_dec list_name_eq_dec loc lck then None else
          bind (sgnlz (expand h' (loc :: lck)) q)
               (fun q' => Some (head_subst s R q'))
        | _ => Some p
        end)
    | Lk_fail => None
    end
  end.

Section Vp_subs.

(* First, prove correctness of vp_subs *)

Variables (h : nat) (lck : list (list name)).
Let sig_subs' := sig_subs (expand h lck).
Let vp_subs' := vp_subs (expand h lck).
Let sgnlz' := sgnlz (expand h lck).

Lemma sig_subs_no_rt : forall sg vp p R,
  vp <> Prt ->
  sig_subs' vp p sg = Some R -> assoc path_beq Prt R = None.
Proof.
  induction sg; simpl; intros.
  Case "Snil".
    fold (sgnlz' p) in *.
    destruct (sgnlz' p).
    SCase "Some".
      simpl in H0. inversion H0.
      destruct vp; simpl; auto.
      elim H; auto.
    SCase "None".
      discriminate.
  Case "SCons".
    case_bind R1.
    case_bind R2.
    inversions H0.
    rewrite <- (IHsg2 _ _ _ H Eq0).
    apply assoc_app2.
    refine (IHsg1 _ _ _ _ Eq).
    intro; discriminate.
Qed.

Lemma vp_subs_no_rt : forall e s n R,
  vp_subs' n e s = Some R -> assoc path_beq Prt R = None.
Proof.
  induction e; destruct s; intros; try discriminate.
  Case "nil".
    inversion H. auto.
  Case "cons".
    simpl in *.
    case_bind R1.
    case_bind R2.
    inversions H.
    rewrite <- (IHe _ _ _ Eq0).
    apply assoc_app2.
    apply* sig_subs_no_rt.
    intro; discriminate.
Qed.

Lemma sig_subs_no_suff : forall sg vp vp' p R,
  path_length vp' < path_length vp ->
  sig_subs' vp p sg = Some R ->
  assoc path_beq vp' R = None.
Proof.
  induction sg; simpl; intros.
  Case "Snil".
    fold (sgnlz' p) in H0.
    destruct (sgnlz' p); simpl in H0; inversion H0.
    simpl.
    case_eq (path_beq vp' vp); intros; auto.
    lets Len: (path_beq_length _ _ H1).
    elimtype False; omega.
  Case "SCons".
    case_bind R1.
    case_bind R2.
    inversions H0.
    rewrite* assoc_app2.
    apply* (IHsg1 (Pdot vp n)).
    simpl; omega.
Qed.

Lemma assoc_sig_subs0 : forall vp p sg R,
  sig_subs' vp p sg = Some R ->
  assoc path_beq vp R = sgnlz' p /\ assoc path_beq vp R <> None.
Proof.
  induction sg; simpl; intros.
    case_bind q.
    inversion H; simpl.
    rewrite path_beq_refl.
    split*.
    intro; discriminate.
  case_bind R1.
  case_bind R2.
  inversion H.
  rewrite* assoc_app2.
  apply* (sig_subs_no_suff sg1 (Pdot vp n)).
Qed.

Lemma sig_subs_no_pref : forall sg vp vp' p ap R,
  path_length vp = path_length vp' ->
  path_beq vp' vp = false ->
  sig_subs' vp p sg = Some R ->
  assoc path_beq (app_path vp' ap) R = None.
Proof.
  induction sg; simpl; intros.
  Case "Snil".
    case_bind q.
    inversions H1.
    simpl.
    destruct ap.
      simpl.
      rewrite* H0.
    case_eq (path_beq (app_path vp' (n::ap)) vp); intros; auto.
    lets Len: (path_beq_length _ _ H1).
    rewrite app_path_length in Len.
    simpl in Len.
    elimtype False; omega.
  Case "SCons".
    case_bind R1.
    case_bind R2.
    inversions H1.
    rewrite assoc_app2.
      apply* (IHsg2 vp vp' p ap).
    destruct ap.
    SCase "nil".
      clear -Eq H H0.
      simpl.
      apply* sig_subs_no_suff.
      simpl; omega.
    SCase "n0::ap".
      simpl.
      apply (IHsg1 (Pdot vp n) (Pdot vp' n0) (Pdot p n) ap); auto.
        simpl. auto.
      simpl.
      rewrite H0.
      destruct (name_eq_dec n0 n); auto.
Qed.

Lemma path_beq_length' : forall p q,
  path_length p <> path_length q -> path_beq p q = false.
Proof.
  intros.
  case_eq (path_beq p q); intros; auto.
  elim H.
  apply* path_beq_length.
Qed.

Lemma assoc_sig_subs_out : forall sg vp p ap R,
  sig_subs' vp p sg = Some R -> 
  ~In_sig ap sg ->
  assoc path_beq (app_path vp ap) R = None.
Proof.
  induction sg; simpl; intros.
  Case "Snil".
    case_bind q.
    inversions H; simpl.
    destruct ap.
      elim H0; auto.
    rewrite* path_beq_length'.
    rewrite app_path_length.
    simpl; omega.
  Case "SCons".
    case_bind R1.
    case_bind R2.
    inversions H.
    destruct ap. elim H0; auto.
    rewrite* assoc_app2.
    simpl.
    destruct (name_eq_dec n0 n).
      subst.
      apply (IHsg1 _ _ _ _ Eq).
      intro; elim H0; auto.
    apply* (sig_subs_no_pref sg1 (Pdot vp n) (Pdot vp n0)).
    simpl.
    destruct* (name_eq_dec n0 n).
Qed.

Lemma assoc_sig_subs : forall sg ap vp p R,
  sig_subs' vp p sg = Some R ->
  In_sig ap sg ->
  assoc path_beq (app_path vp ap) R = sgnlz (expand h lck) (app_path p ap) /\
  assoc path_beq (app_path vp ap) R <> None.
Proof.
  induction sg; simpl; introv Eq Hap.
    inversions Hap.
    case_bind q.
    inversions Eq.
    simpl.
    rewrite path_beq_refl. intuition. discriminate.
  destruct ap.
  Case "nil".
    simpl.
    apply assoc_sig_subs0 with (sg:=Scons n sg1 sg2).
    assumption.
  Case "n0::ap".
    case_bind R1.
    case_bind R2.
    inversions Eq.
    destruct (name_eq_dec n0 n).
    SCase "n0=n".
      subst.
      destruct (In_sig_dec ap sg1).
        simpl.
        destruct (IHsg1 _ _ _ _ Eq0 i).
        rewrite assoc_app.
        case_rewrite HR1 (assoc path_beq (app_path (Pdot vp n) ap) R1).
          auto.
        elim H0; auto.
      rewrite assoc_app2.
        apply* (IHsg2 (n::ap)).
        inversions* Hap.
      apply (assoc_sig_subs_out _ _ Eq0 n0).
    SCase "n0<>n".
      rewrite assoc_app2.
        apply* (IHsg2 (n0::ap)).
        inversions* Hap.
      simpl.
      apply* (sig_subs_no_pref sg1 (Pdot vp n)).
      simpl.
      destruct* (name_eq_dec n0 n).
 Qed.

Lemma assoc_vp_subs : forall k sg p ap e s R,
  vp_subs' 0 e s = Some R ->
  length e = length s ->
  is_nth k p s -> is_nth k sg e ->
  In_sig ap sg ->
  assoc path_beq (app_path (Pvar k sg) ap) R = sgnlz' (app_path p ap) /\
  sgnlz' (app_path p ap) <> None.
Proof.
  introv Eq He Hp Hsg Hap.
  revert Eq.
  replace k with (k+0) by auto with arith.
  generalize 0.
  gen e R.
  induction Hp; destruct e; simpl; intros; try discriminate.
  Case "0".
    case_bind R1.
    case_bind R2.
    inverts Eq.
    inverts He.
    inverts Hsg.
    destruct* (@assoc_sig_subs s ap (Pvar n s) a R1).
      fold (sgnlz' (app_path a ap)) in *.
      rewrite assoc_app, H.
      destruct* (sgnlz' (app_path a ap)).
      intuition; discriminate.
  Case "S n".
    case_bind R1.
    case_bind R2.
    inverts Eq.
    inverts He.
    inverts Hsg.
    rewrite assoc_app2.
      simpl.
      rewrite plus_n_Sm.
      apply* IHHp.
    refine (sig_subs_no_pref _ _ _ _ _ _ _ Eq0).
      simpl; auto.
    simpl.
    destruct n0. auto.
    case_eq (nat_beq (n + S n0) n0); intros.
      assert (Hn: n + S n0 <> n0) by omega.
      elim Hn; apply* nat_dec_bl.
    auto.
Qed.

End Vp_subs.

(* Monotonicity wrt to h and lck *)

Section Monotone.
Variables (h h':nat) (lck lck':list(list name)).
Variable IHh: forall p q, 
  expand h lck p = Some q -> expand h' lck' p = Some q.
Variable Hh: h <= h'.
Variable Hlck: incl lck' lck.

Lemma sgnlz_monotone0 : forall p q,
  sgnlz (expand h lck) p = Some q ->
  sgnlz (expand h' lck') p = Some q.
Proof.
  induction p; simpl; intros; auto*.
  Case "Pdot".
    case_bind p'.
    rewrite* (IHp p').
    simpls*.
  Case "Papp".
    case_bind p1'.
    rewrite* (IHp1 p1').
    simpls*.
Qed.

Lemma sig_subs_monotone : forall sg vp p R,
  sig_subs (expand h lck) vp p sg = Some R ->
  sig_subs (expand h' lck') vp p sg = Some R.
Proof.
  induction sg; simpl; intros.
  SSCase "Snil".
    case_bind p'.
    rewrite* (sgnlz_monotone0 _ Eq).
  SSCase "Scons".
    case_bind R1.
    case_bind R2.
    rewrite (IHsg1 _ _ _ Eq).
    rewrite (IHsg2 _ _ _ Eq0).
    assumption.
Qed.

Lemma vp_subs_monotone: forall e s n R,
  vp_subs (expand h lck) n e s = Some R -> length e = length s ->
  vp_subs (expand h' lck') n e s = Some R.
Proof.
  induction e; destruct s; simpl; intros; try discriminate; auto.
  Case "a::e".
  case_bind R1.
  case_bind R2.
  rewrite sig_subs_monotone with (R:=R1) by auto.
  rewrite IHe with (R:=R2) by auto.
  assumption.
Qed.
End Monotone.

Lemma expand_monotone : forall h lck h' lck',
  h <= h' -> incl lck' lck ->
  forall p q, expand h lck p = Some q -> expand h' lck' p = Some q.
Proof.
  induction h.
  Case "0".
    simpl; intros. discriminate.
  Case "S h".
  introv Hh Hlck; intros.
  destruct h'. elimtype False; omega.
  assert (Hh': h <= h') by omega.
  clear Hh.
  simpl in *.
  case_rewrite lkp (lookup_dec p).
    case_bind R.
    rewrite* (vp_subs_monotone _ _ _ _ (IHh _ _ _ Hh' Hlck) e s 0 Eq).
    simpl.
    destruct e0; auto.
    destruct (in_dec list_name_eq_dec l lck); try discriminate.
    destruct (in_dec list_name_eq_dec l lck').
      elim n.
      apply* Hlck.
    case_bind p0'.
    assert (sgnlz (expand h' (l :: lck')) p0 = Some p0').
      apply* (@sgnlz_monotone0 h h' (l :: lck) (l :: lck')).
      intros; apply* IHh.
      intros x; simpl; intros.
      destruct* H1.
    rewrite* H0.
  destruct* (In_sig_dec l s).
Qed.

Theorem sgnlz_monotone : forall h lck h' lck' p q,
  h <= h' -> incl lck' lck ->
  sgnlz (expand h lck) p = Some q -> sgnlz (expand h' lck') p = Some q.
Proof.
  intros.
  apply* (@sgnlz_monotone0 h h' lck).
  apply* expand_monotone.
Qed.

(* Properties of lookup_dec and rooted_path *)

Lemma lookup_dec_subst : forall p e s d ap vs,
  lookup_dec p = Lk_ok e s d ap ->
  lookup_dec (subst vs p) = Lk_ok e (map (subst vs) s) d ap.
Proof.
  induction p; simpl; intros; try solve [inverts H; auto].
  Case "Pdot".
    case_rewrite lkp (lookup_dec p).
    destruct e1; try discriminate.
    case_rewrite Hass (assoc name_beq n l0).
    rewrite (IHp _ _ _ _ _ (refl_equal _)).
    rewrite Hass.
    inverts* H.
  Case "Papp".
    case_rewrite lkp1 (lookup_dec p1).
    destruct e1; try discriminate.
    inverts* H.
    rewrite (IHp1 _ _ _ _ _ (refl_equal _)).
    reflexivity.
Qed.

Lemma subst_rooted : forall s p,
  rooted_path p = true -> rooted_path (subst s p) = true.
Proof.
  induction p; simpl; intros; auto. discriminate.
Qed.

Lemma lk_var_not_rooted : forall p x s l,
  lookup_dec p = Lk_var x s l -> rooted_path p = false.
Proof.
  intros until s.
  induction p; simpl; intros; try discriminate; auto.
  Case "Pdot".
    case_rewrite lkp (lookup_dec p).
      destruct e0; try discriminate.
      destruct (assoc name_beq n l1); discriminate.
    inverts* H.
  Case "Papp".
    destruct (lookup_dec p1); try discriminate.
    destruct e0; discriminate.
Qed.

Lemma lookup_dec_var : forall p n s ap,
  lookup_dec p = Lk_var n s ap -> p = app_path (Pvar n s) ap.
Proof.
  intros until s; induction p; simpl; intros; try discriminate.
  Case "Pvar". inverts* H.
  Case "Pdot".
    case_rewrite lkp (lookup_dec p).
    destruct e0; try discriminate.
      destruct (assoc name_beq n0 l0); discriminate.
    inverts H.
    rewrite app_path_app.
    replace (Pdot p n0) with (app_path p (n0 :: nil)) by reflexivity.
    f_equal.
    auto*.
  Case "Papp".
    destruct (lookup_dec p1); try discriminate.
    destruct e0; discriminate.
Qed.

(* Closed app_paths *)

Lemma closed_var_is_nth : forall e n s l,
  closed e (app_path (Pvar n s) l) -> is_nth n s e.
Proof.
  intros; induction l using rev_ind.
  inverts* H.
  rewrite app_path_app in H.
  simpl in H.
  inverts* H.
Qed.

Lemma closed_nth : forall e x l,
  list_forall (closed e) l -> closed e (nth x l Prt).
Proof.
  intros; gen x.
  induction H; simpl; intros. destruct* x. destruct* x0.
Qed.

Lemma closed_app_path : forall e p ap,
  closed e p -> closed e (app_path p ap).
Proof.
  induction ap using rev_ind; simpl; intros; auto.
  rewrite app_path_app; simpl.
  auto.
Qed.

(* Properties of the result of semi-ground normalization *)

Inductive sgnlz_res : path -> Prop :=
  | Sgnlz_ok : forall p e s d ,
    lookup P p e s d -> nonpath d -> sgnlz_res p
  | Sgnlz_var : forall p x s l,
    lookup_dec p = Lk_var x s l -> In_sig l s -> sgnlz_res p.
Hint Constructors sgnlz_res.

Section Sgnlz_res.
Variables (h:nat) (lck:list(list name)) (e:env).
Hypothesis IHh : forall p q,
  expand h lck p = Some q -> closed e p -> sgnlz_res q /\ closed e q.

Lemma sgnlz_res_ok : forall p q,
  sgnlz (expand h lck) p = Some q -> closed e p ->
  sgnlz_res q /\ closed e q.
Proof.
  induction p; simpl; intros.
  Case "Prt".
    apply* IHh.
  Case "Pvar".
    apply* IHh.
  Case "Pdot".
    case_bind p'.
    inverts H0.
    destruct* (IHp p').
    apply* IHh.
  Case "Papp".
    case_bind p1'.
    inverts H0.
    destruct* (IHp1 p1').
    apply* IHh.
Qed.
End Sgnlz_res.

Lemma expand_res_ok : forall h lck e p q,
  expand h lck p = Some q -> closed e p -> sgnlz_res q /\ closed e q.
Proof.
  induction h; simpl; intros until q; intros Hp Hcl.
    discriminate.
  case_rewrite lkp (lookup_dec p).
  Case "Lk_ok".
    case_bind R.
    lets Cls: (lookup_args_closed (lookup_sound _ lkp) Hcl).
    destruct e1; try solve [inverts* Hp].
    destruct (in_dec list_name_eq_dec l lck).
      discriminate.
    case_bind p0'.
    unfold head_subst in Hp.
    destruct* (sgnlz_res_ok _ _ (IHh _ e0) Eq0) as [Hres Hcl'].
       apply* lookup_closed.
    inverts Hres.
    SCase "Sgnlz_ok".
      rewrite (lookup_rooted H) in Hp.
      inverts Hp.
      split.
        eapply Sgnlz_ok.
          apply* lookup_subst.
        assumption.
      apply* (subst_closed Cls Hcl').
    SCase "Sgnlz_var".
      rewrite (lk_var_not_rooted _ H) in Hp.
      rewrite (lookup_dec_var _ H) in *.
      lets x_nth: (closed_var_is_nth _ _ _ Hcl').
      lets Hlen: (lookup_length (lookup_sound _ lkp)).
      destruct* (@assoc_vp_subs h lck x s0 (nth x s Prt) l1 e0 s R).
        apply* nth_is_nth.
        rewrite <- Hlen.
        apply* is_nth_lt.
      rewrite H1 in Hp.
      case_rewrite Hass (sgnlz (expand h lck) (app_path (nth x s Prt) l1));
        auto*.
      inverts Hp.
      destruct* (sgnlz_res_ok _ _ (IHh _ e) Hass).
      apply closed_app_path.
      apply* closed_nth.
  Case "Lk_var".
    destruct (In_sig_dec l s); try discriminate.
    inverts Hp.
    splits*.
Qed.

(* Proof of correctness wrt to safe reduction *)

Section Correct.
Variables (h:nat) (lck:list(list name)).
Hypothesis Hexpand : forall p q,
  expand h lck p = Some q -> red P p q.

Lemma sgnlz_correct : forall p q,
  sgnlz (expand h lck) p = Some q -> red P p q.
Proof.
  induction p; simpl; intros; auto*.
  Case "Pdot".
    case_bind p'.
    apply* r_dot.
  Case "Papp".
    case_bind p1'.
    apply* r_app.
Qed.

Lemma sig_subs_correct : forall sg vp p R,
  sig_subs (expand h lck) vp p sg = Some R -> safe P p sg.
Proof.
  induction sg; simpl; intros.
  Case "Snil".
    case_bind p'.
    apply safe_nil with p'.
    apply* sgnlz_correct.
  Case "Scons".
    case_bind R1.
    case_bind R2.
    inverts H.
    apply (@safe_cons _ p n sg1 sg2); auto*.
Qed.

Lemma vp_subs_correct : forall e s n R,
  vp_subs (expand h lck) n e s = Some R ->
  length e = length s ->
  safe_subs P s e.
Proof.
  induction e; destruct s; simpl; intros; try discriminate.
  Case "nil". constructor.
  Case "cons".
    case_bind R1.
    case_bind R2.
    constructor.
      apply* sig_subs_correct.
    inverts* H0.
Qed.
End Correct.


Theorem expand_correct : forall h lck p q,
  expand h lck p = Some q -> red P p q.
Proof.
Hint Resolve vp_subs_correct lookup_length.
  induction h; intros.
    discriminate.
  simpl in H.
  case_rewrite lkp (lookup_dec p).
  Case "Lk_ok".
    case_bind R.
    destruct e0.
    lets Hlen: (lookup_length (lookup_sound _ lkp)).
    SCase "Fun".
      inverts H.
      apply* r_src.
    SCase "Mod".
      inverts H.
      apply* r_src.
    SCase "Abb".
      destruct (in_dec list_name_eq_dec l lck). discriminate.
      case_bind p0'.
      inverts H.
      apply* r_exp.
      assert (Hlck: incl lck (l::lck)).
        intros x Hx; right*.
      assert (sgnlz (expand h lck) p0 = Some p0') by apply* sgnlz_monotone.
      lets Red: (sgnlz_correct _ _ (IHh lck) _ H).
      assert (Safe: safe_subs P s e) by auto*.
      destruct (substitution P_ok (safe_nil Red) Safe) as [Safe' Hred].
        apply* lookup_closed.
      specialize (Hred _ Red).
      unfold head_subst.
      case_rewrite HR (rooted_path p0').
        auto*.
      destruct Hred as [r [Rp0 Rp0']].
      destruct (sgnlz_res_ok _ _ (@expand_res_ok h lck e) H).
        apply* lookup_closed.
      inverts H0.
        rewrite (lookup_rooted H2) in HR. discriminate.
      rewrite (lookup_dec_var _ H2) in *.
      lets Nth: closed_var_is_nth H1.
      assert (Lt: x < length s).
        assert (Hlen: length e = length s) by auto*.
        rewrite <- Hlen.
        apply* is_nth_lt.
      destruct* (@assoc_vp_subs h lck x s0 (nth x s Prt) l1 e s _ Eq).
        apply* nth_is_nth.
      rewrite H0.
      case_rewrite Hx (sgnlz (expand h lck) (app_path (nth x s Prt) l1)).
        lets Red': sgnlz_correct IHh Hx.
        rewrite subst_vp in Rp0'.
        simpl in Rp0'.
        lets Nth': (nth_is_nth _ Prt Lt).
        rewrite <- (is_nth_nth (Pvar x s0) Nth') in Red'.
        rewrite (red_unique P_ok Red' Rp0').
        assumption.
      elim H4; auto.
  Case "Lk_var".
    destruct (In_sig_dec l s); try discriminate.
    inverts H.
    rewrite (lookup_dec_var _ lkp).
    apply r_vp.
    auto.
Qed.

Definition correctness h lck := sgnlz_correct _ _ (@expand_correct h lck).


(* Lemmas for postponement *)

Lemma sgnlz_nonpath : forall h lck p e s d ap,
  lookup_dec p = Lk_ok e s d ap -> nonpath d ->
  sgnlz (expand h lck) p <> None -> sgnlz (expand h lck) p = Some p.
Proof.
  induction p; simpl; intros.
  Case "Prt".
    inverts H.
    subst d.
    destruct* h.
    simpl in *.
    destruct* P.
    inverts H0.
  Case "Pvar".
    inverts H.
  Case "Pdot".
    case_rewrite lkp (lookup_dec p).
    destruct e1; try discriminate.
    case_rewrite Hass (assoc name_beq n l0).
    inverts H.
    case_bind p'; auto*.
    simpl.
    assert (Some p' = Some p).
      apply* (IHp e s (Mod l0) l). intro; discriminate.
    clear IHp; inverts H.
    destruct* h.
    simpl in *.
    rewrite lkp, Hass in *.
    case_bind R; auto*.
    case_rewrite Hd d; auto*.
    simpl in *.
    inverts H0.
  Case "Papp".
    clear IHp2.
    case_rewrite lkp1 (lookup_dec p1).
    destruct e1; try discriminate.
    inverts H.
    case_bind p1'; auto*.
    simpl.
    assert (Some p1' = Some p1).
      apply* (IHp1 e0 s0 (Fun s1 d) ap). intro; discriminate.
    clear IHp1; inverts H.
    destruct* h.
    simpl in *.
    rewrite lkp1 in *.
    case_bind R; auto*.
    case_rewrite Hd d; auto*.
    simpl in *.
    inverts H0.
Qed.

Lemma expand_nonpath : forall h lck p e s d ap R,
  lookup_dec p = Lk_ok e s d ap -> nonpath d ->
  vp_subs (expand h lck) 0 e s = Some R ->
  expand (S h) lck p = Some p.
Proof.
  simpl; introv lkp Hnp HR.
  rewrite lkp.
  rewrite HR.
  simpl.
  inverts* Hnp.
Qed.

Lemma In_sig_impl : forall l l' sg, In_sig (l++l') sg -> In_sig l sg.
Proof.
  intros.
  gen l; induction sg; simpl; intros.
    destruct l; destruct l'; inverts* H.
  destruct* l.
  inverts* H.
Qed.

Lemma sig_subs_fail_impl : forall h lck v v' p sg,
   sig_subs (expand h lck) v p sg = None ->
   sig_subs (expand h lck) v' p sg = None.
Proof.
  intros.
  gen v v' p; induction sg; simpl; intros.
    case_bind p'; auto.
  case_bind R1.
    case_bind R2.
    destruct* (sig_subs (expand h lck) (Pdot v' n) (Pdot p n) sg1).
    simpl.
    rewrite* (IHsg2 _ v' _ Eq0).
  rewrite* (IHsg1 _ (Pdot v' n) _ Eq).
Qed.

Lemma vp_subs_fail_impl : forall h lck n e s,
  vp_subs (expand h lck) n e s = None ->
  vp_subs (expand h lck) (S n) e s = None.
Proof.
  intros.
  gen n s; induction e; destruct s; simpl; intros; try discriminate; auto.
  case_bind R1.
  case_bind R2.
    rewrite (IHe _ _ Eq0).
    destruct* (sig_subs (expand h lck) (Pvar (S n) a) p a).
  rewrite (sig_subs_fail_impl _ _ _ _ _ _ Eq).
  auto.
Qed.


Lemma expand_fail : forall h lck p1,
  expand h lck p1 = None ->
  (forall m, expand h lck (Pdot p1 m) = None) /\
  (forall p2, expand h lck (Papp p1 p2) = None).
Proof.
  intros.
  destruct h; simpl in *.
    split*.
  split; intros; case_rewrite lkp1 (lookup_dec p1); auto*.
      case_bind R; destruct e0; try discriminate; auto.
      case_eq (assoc name_beq m l0); introv Hass; auto.
      rewrite* Eq.
    destruct* (In_sig_dec (l++m::nil) s).
    destruct (In_sig_dec l s); try discriminate.
    elim n; apply* In_sig_impl.
  case_bind R; destruct e0; try discriminate; auto.
  simpl.
  rewrite (vp_subs_fail_impl _ _ _ _ _ Eq).
  destruct* (sig_subs (expand h lck) (Pvar 0 s0) p2 s0).
Qed.
  
Lemma lookup_sgnlz : forall h lck p e s d,
  lookup P p e s d ->
  sgnlz (expand h lck) p = expand h lck p.
Proof.
  intros.
  destruct (lookup_complete H) as [ap lkp].
  clear H; gen e s d ap.
  induction p; simpl; intros; auto.
  Case "Pdot".
    case_rewrite lkp' (lookup_dec p).
    destruct e1; try discriminate.
    case_rewrite Hass (assoc name_beq n l0).
    case_eq (sgnlz (expand h lck) p); introv Hp.
      simpl.
      assert (nonpath (Mod l0)) by constructor.
      rewrite* (sgnlz_nonpath h lck _ lkp') in Hp.
        inverts* Hp.
      intro Hp'; rewrite Hp' in Hp; discriminate.
    rewrite* (IHp e0 s0 (Mod l0) l) in Hp.
    destruct (expand_fail _ _ _ Hp).
    rewrite* H.
  Case "Papp".
    case_rewrite lkp1 (lookup_dec p1).
    destruct e1; try discriminate.
    case_eq (sgnlz (expand h lck) p1); introv Hp1.
      simpl.
      assert (nonpath (Fun s1 e1)) by constructor.
      rewrite* (sgnlz_nonpath h lck _ lkp1) in Hp1.
        inverts* Hp1.
      intro Hp'; rewrite Hp' in Hp1; discriminate.
    rewrite* (IHp1 e0 s0 (Fun s1 e1) l) in Hp1.
    destruct (expand_fail _ _ _ Hp1).
    rewrite* H0.
Qed.

Lemma lookup_sgnlz_app : forall h lck s h1 p1 p2 q e0 s0 d,
  expand h lck (Papp p1 p2) = Some q ->
  lookup P p1 e0 s0 d ->
  sgnlz (expand h1 lck) (subst s (Papp p1 p2)) = 
  expand h1 lck (subst s (Papp p1 p2)).
Proof.
  introv Hq Hlk.
  destruct h; simpl in Hq. discriminate.
  destruct (lookup_complete Hlk) as [ap lkp].
  rewrite lkp in Hq.
  destruct d; try discriminate.
  assert (lookup P (Papp p1 p2) (s1::e0) (p2::s0) d).
    constructor.
    apply (lookup_sound _ lkp).
  lets lkp': (lookup_subst s H).
  apply (lookup_sgnlz _ _ lkp').
Qed.

Lemma lookup_sgnlz_dot : forall h lck s h1 p' n q e0 s0 d,
  expand h lck (Pdot p' n) = Some q ->
  lookup P p' e0 s0 d ->
  sgnlz (expand h1 lck) (subst s (Pdot p' n)) = 
  expand h1 lck (subst s (Pdot p' n)).
Proof.
  introv Hq Hlk.
  destruct h; simpl in Hq. discriminate.
  destruct (lookup_complete Hlk) as [ap lkp].
  rewrite lkp in Hq.
  destruct d; try discriminate.
  case_rewrite Hass (assoc name_beq n l).
  assert (lookup P (Pdot p' n) e0 s0 e).
    econstructor; auto*.
  lets lkp': (lookup_subst s H).
  apply (lookup_sgnlz _ _ lkp').
Qed.

Require Import Max.
Hint Immediate le_max_l le_max_r.

(* Proof of postponement *)

Section Postponement.

Variables (h':nat) (lck':list(list name)).
Variables (e:env) (s:list path) (R:list (path*path)).
Hypothesis HR: vp_subs (expand h' lck') 0 e s = Some R.
Hypothesis Hlen: length e = length s.
Variable lck : list (list name).
Hypothesis Hlck: incl lck lck'.

Lemma vp_postpone: forall v s0 l q,
  lookup_dec q = Lk_var v s0 l ->
  In_sig l s0 ->
  closed e q ->
  sgnlz (expand h' lck) (subst s q) = Some (head_subst s R q).
Proof.
  introv lkp Hsig Hcl.
  rewrite (lookup_dec_var _ lkp) in *.
  rewrite subst_vp.
  lets Hve: (closed_var_is_nth _ _ _ Hcl).
  unfold head_subst.
  rewrite rooted_vp.
  lets Lt: (is_nth_lt Hve).
  rewrite Hlen in Lt.
  lets Hvs: (nth_is_nth _ Prt Lt).
  destruct (assoc_vp_subs _ _ HR Hlen Hvs Hve Hsig).
  simpl subst.
  rewrite (is_nth_nth _ Hvs).
  rewrite H.
  case_rewrite Hap (sgnlz (expand h' lck') (app_path (nth v s Prt) l)).
    apply* sgnlz_monotone.
  elim H0; auto.
Qed.

Variable h : nat.
Hypothesis Hexpand : forall p q,
  expand h lck p = Some q -> closed e p ->
  exists h,
    sgnlz (expand h lck) (subst s p) = Some (head_subst s R q) /\
    sgnlz (expand h lck) (subst s q) = Some (head_subst s R q).

Lemma sgnlz_weaken : forall h h' lck p q,
  sgnlz (expand h' lck) p = Some q -> h' <= h ->
  sgnlz (expand h lck) p = Some q.
Proof.
  clear.
  intros.
  assert (incl lck lck). intros x Hx; auto.
  apply* (@sgnlz_monotone h').
Qed.

Lemma sgnlz_postpone : forall p q,
  sgnlz (expand h lck) p = Some q ->
  closed e p ->
  exists h,
    sgnlz (expand h lck) (subst s p) = Some (head_subst s R q) /\
    sgnlz (expand h lck) (subst s q) = Some (head_subst s R q).
Proof.
  induction p; intros until q; intros Hq Hcl.
  Case "Prt".
    apply* (@Hexpand Prt q).
  Case "Pvar".
    apply* (@Hexpand (Pvar v s0) q).
  Case "Pdot".
    simpl in *.
    case_bind p'.
    inverts Hcl.
    destruct* (sgnlz_res_ok _ _ (expand_res_ok h lck (e:=e)) Eq) as [Hres Hcl].
    destruct* (IHp p') as [h'' [Hsp Hsp']].
    destruct* (Hexpand Hq) as [h2 [Hspn Hsq]].
    pose (h1 := max h' (max h'' h2)).
    exists h1.
    assert (Hh': h' <= h1) by (unfold h1; auto).
    assert (Hh'': h'' <= h1) by (unfold h1; eauto with arith).
    assert (Hh2: h2 <= h1) by (unfold h1; eauto with arith).
    rewrite* <- (sgnlz_weaken (h:=h1) _ _ Hsq).
    split*.
    rewrite* (sgnlz_weaken (h:=h1) _ _ Hsp).
    simpl.
    inverts Hres.
    SCase "Sgnlz_ok".
      rewrite* (sgnlz_weaken (h:=h1) _ _ Hsq).
      unfold head_subst at 1.
      rewrite (lookup_rooted H).
      replace (Pdot (subst s p') n) with (subst s (Pdot p' n)) by auto.
      rewrite* <- (sgnlz_weaken (h:=h1) _ _ Hspn).
      symmetry.
      apply* lookup_sgnlz_dot.
    SCase "Sgnlz_var".
    clear -Hq HR Hlen Hlck Hcl H H1 Hh'.
    assert (q = Pdot p' n).
      destruct h; simpl in Hq. discriminate.
      rewrite H in Hq.
      destruct (In_sig_dec (l0 ++ n :: nil) s0); try discriminate.
      inverts* Hq.
    subst q.
    rewrite (lookup_dec_var _ H) in *.
    lets Hve: (closed_var_is_nth _ _ _ Hcl).
    unfold head_subst.
    rewrite rooted_vp.
    lets Lt: (is_nth_lt Hve).
    rewrite Hlen in Lt.
    lets Hvs: (nth_is_nth _ Prt Lt).
    destruct (assoc_vp_subs _ _ HR Hlen Hvs Hve H1).
    rewrite H0.
    case_rewrite Hx (sgnlz (expand h' lck') (app_path (nth x s Prt) l0));
      auto*.
    simpl.
    rewrite subst_vp.
    simpl.
    rewrite (is_nth_nth _ Hvs).
    rewrite* (sgnlz_monotone _ Hh' Hlck Hx).
  Case "Papp".
    simpl in *.
    case_bind p'.
    inverts Hcl.
    destruct* (sgnlz_res_ok _ _ (expand_res_ok h lck (e:=e)) Eq) as [Hres Hcl].
    destruct* (IHp1 p') as [h'' [Hsp Hsp']].
    destruct* (Hexpand Hq) as [h2 [Hspn Hsq]].
    pose (h1 := Max.max h' (Max.max h'' h2)).
    exists h1.
    assert (Hh': h' <= h1) by (unfold h1; auto).
    assert (Hh'': h'' <= h1) by (unfold h1; eauto with arith).
    assert (Hh2: h2 <= h1) by (unfold h1; eauto with arith).
    rewrite* <- (sgnlz_weaken (h:=h1) _ _ Hsq).
    split*.
    rewrite* (sgnlz_weaken (h:=h1) _ _ Hsp).
    simpl.
    inverts Hres.
    SCase "Sgnlz_ok".
      rewrite* (sgnlz_weaken (h:=h1) _ _ Hsq).
      unfold head_subst at 1.
      rewrite (lookup_rooted H).
      replace (Papp (subst s p') (subst s p2)) with (subst s (Papp p' p2))
        by auto.
      rewrite* <- (sgnlz_weaken (h:=h1) _ _ Hspn).
      symmetry.
      apply* lookup_sgnlz_app.
    SCase "Sgnlz_var".
    elimtype False.
    clear -H Hq.
    destruct h; simpl in Hq. discriminate.
    rewrite H in Hq.
    discriminate.
Qed.

Lemma sig_subs_postpone: forall sg p q R1,
  sig_subs (expand h lck) p q sg = Some R1 -> closed e q ->
  exists h1,
    sig_subs (expand h1 lck) p (subst s q) sg =
    Some (map_snd (head_subst s R) R1).
Proof.
  clear Hexpand.
  intros until sg.
  induction sg; simpl; intros.
    case_bind x'.
    inverts H.
    destruct* (sgnlz_postpone Eq) as [h'' [Hsq Hsx']].
    exists h''.
    rewrite* Hsq.
  case_bind R2.
  case_bind R3.
  inverts H.
  destruct* (IHsg1 _ _ _ Eq) as [h1 HL1].
  destruct* (IHsg2 _ _ _ Eq0) as [h2 HL2].
  exists (max h1 h2).
  simpl subst in HL1.
  rewrite (sig_subs_monotone _ _ _ _ 
          (expand_monotone (le_max_l h1 h2) (incl_refl lck)) _ _ _ HL1).
  rewrite (sig_subs_monotone _ _ _ _ 
          (expand_monotone (le_max_r h1 h2) (incl_refl lck)) _ _ _ HL2).
  simpl.
  unfold map_snd; rewrite* map_app.
Qed.

Lemma vp_subs_postpone : forall n e' s' R',
  vp_subs (expand h lck) n e' s' = Some R' ->
  list_forall (closed e) s' ->
  length e' = length s' ->
  exists h1,
    vp_subs (expand h1 lck) n e' (map (subst s) s') =
    Some (map_snd (head_subst s R) R').
Proof.
  introv HR' Hcl Hlen'.
  gen n e' R'; induction Hcl; destruct e'; simpl; intros; try discriminate.
    inverts* HR'.
  inverts Hlen'.
  case_bind R1.
  case_bind R2.
  destruct* (IHHcl _ _ H1 _ Eq0) as [h2 HL2].
  destruct* (sig_subs_postpone _ _ Eq) as [h1 HL1].
  exists (max h1 h2).
  rewrite (sig_subs_monotone _ _ _ _ 
    (expand_monotone (le_max_l h1 h2) (incl_refl lck)) _ _ _ HL1).
  rewrite (vp_subs_monotone _ _ _ _ 
    (expand_monotone (le_max_r h1 h2) (incl_refl lck)) _ _ _ HL2).
  simpl.
  inverts HR'.
  unfold map_snd; rewrite* map_app.
  rewrite* map_length.
Qed.

End Postponement.

Lemma expand_postpone : forall h lck h' lck' e s R p q,
  vp_subs (expand h' lck') 0 e s = Some R ->
  length e = length s -> incl lck lck' ->
  expand h lck p = Some q -> closed e p ->
  exists h'',
    sgnlz (expand h'' lck) (subst s p) = Some (head_subst s R q) /\
    sgnlz (expand h'' lck) (subst s q) = Some (head_subst s R q).
Proof.
Hint Resolve lookup_args_closed.
  induction h.
    intros; discriminate.
  intros until q; intros HR Hlen Hlck Hq Hcl.
  simpl in Hq.
  case_rewrite lkp (lookup_dec p).
  Case "Lk_ok".
    case_rewrite Eqvp (vp_subs (expand h lck) 0 e0 s0).
    rename l0 into R1.
    simpl in Hq.
    lets lksp: (lookup_dec_subst _ s lkp).
    pose (L1:=map_snd (head_subst s R) R1).
    assert (exists h1,
            vp_subs (expand h1 lck) 0 e0 (map (subst s) s0) = Some L1).
      apply* vp_subs_postpone.
    destruct H as [h1 HL1].
    destruct e1.
    SCase "Fun".
      inverts Hq.
      exists (S h1).
      rewrite (lookup_sgnlz _ _ (lookup_sound _ lksp)).
      simpl.
      rewrite lksp.
      rewrite HL1.
      simpl.
      unfold head_subst.
      rewrite* (lookup_rooted (lookup_sound _ lkp)).
    SCase "Mod".
      inverts Hq.
      exists (S h1).
      rewrite (lookup_sgnlz _ _ (lookup_sound _ lksp)).
      simpl.
      rewrite lksp.
      rewrite HL1.
      simpl.
      unfold head_subst.
      rewrite* (lookup_rooted (lookup_sound _ lkp)).
    SCase "Abb".
      destruct (in_dec list_name_eq_dec l lck). discriminate.
      case_bind p0'.
      inverts Hq.
      lets Hcl0: (lookup_closed P_ok (lookup_sound _ lkp)).
      lets Len: (lookup_length (lookup_sound _ lkp)).
      destruct* (sgnlz_res_ok _ _ (@expand_res_ok h (l::lck) e0) Eq)
        as [Hres Hcl'].
      inverts Hres.
      SSCase "Lk_ok".
        pose (s':= map (subst s) s0).
        assert (exists h,
          sgnlz(expand h lck)(subst s' p0) = Some(head_subst s' L1 p0')
          /\
          sgnlz(expand h lck)(subst s' p0') = Some(head_subst s' L1 p0')).
          subst s'; apply* sgnlz_postpone.
          apply incl_refl.
          intros; apply* IHh.
          apply incl_refl.
          assert (incl lck (l::lck)). intros x Hx; right*.
          apply* (@sgnlz_monotone h).
        destruct H1 as [h2 [Hsp0 Hsp0']].
        exists (S (max h (max h1 h2))).
        assert (Hh: h <= max h (max h1 h2)) by auto*.
        assert (Hh1: h1 <= max h (max h1 h2)) by eauto with arith.
        assert (Hh2: h2 <= S (max h (max h1 h2))) by eauto with arith.
        rewrite (lookup_sgnlz _ _ (lookup_sound _ lksp)).
        simpl expand at 1.
        rewrite lksp.
        rewrite (vp_subs_monotone _ _ _ _
                  (expand_monotone Hh1 (incl_refl lck)) _ _ _ HL1).
        destruct* (in_dec list_name_eq_dec l lck).
        clear n0.
        simpl bind.
        rewrite (sgnlz_monotone _ Hh (incl_refl _) Eq).
        simpl bind.
        unfold head_subst at 1 3 4 6.
        rewrite (lookup_rooted H).
        unfold head_subst.
        rewrite (subst_rooted s0 _ (lookup_rooted H)).
        rewrite* (subst_map s s0 Hcl').
        split*.
        rewrite* <- (subst_map s s0 Hcl').
        fold s'.
        rewrite (sgnlz_monotone _ Hh2 (incl_refl _) Hsp0').
        unfold head_subst.
        rewrite* (lookup_rooted H).
        rewrite* map_length.
      SSCase "Lk_var".
        lets Cls: (lookup_args_closed (e:=e0) (lookup_sound _ lkp) Hcl).
        unfold head_subst at 2 3 5.
        rewrite (lk_var_not_rooted _ H).
        rewrite (lookup_dec_var _ H) in *.
        lets: (closed_app_inv _ _ Hcl').
        inverts H1.
        destruct* (@assoc_vp_subs h lck x s1 (nth x s0 Prt) l1 e0 s0 _ Eqvp).
          apply nth_is_nth.
          rewrite <- Len.
          apply* is_nth_lt.
        rewrite <- (lookup_dec_var _ H) in *.
        rewrite <- H1 in H2.
        case_rewrite Hass (assoc path_beq p0' R1); auto*.
        clear H2.
        set (p1a := app_path (nth x s0 Prt) l1) in *.
        assert (exists h,
          sgnlz (expand h lck) (subst s p1a) = Some (head_subst s R p1) /\
          sgnlz (expand h lck) (subst s p1) = Some (head_subst s R p1)).
          apply* sgnlz_postpone.
          unfold p1a.
          apply closed_app_path.
          apply* closed_nth.
        destruct H2 as [h2 [Hp1a Hp1]].
        pose (h0 := max (max h1 h2) (max h h')). 
        exists (S h0).
        rewrite (lookup_sgnlz _ _ (lookup_sound _ lksp)).
        simpl expand at 1.
        rewrite lksp.
        assert (Hh1: h1 <= h0) by (unfold h0; eauto with arith).
        rewrite (vp_subs_monotone _ _ _ _ (expand_monotone Hh1 (incl_refl lck))
                 _ _ _ HL1).
        destruct* (in_dec list_name_eq_dec l lck).
        clear n0.
        simpl bind.
        assert (Hh: h <= h0) by (unfold h0; eauto with arith).
        assert (Hlck': incl lck (l::lck)) by (intros a Ha; right*).
        rewrite (sgnlz_monotone _ Hh (incl_refl _) Eq).
        simpl bind.
        unfold head_subst at 1.
        rewrite (lk_var_not_rooted _ H).
        unfold L1; rewrite map_assoc.
        rewrite Hass.
        split*.
        assert (Hh2: h2 <= S h0) by (unfold h0; eauto with arith).
        apply* (sgnlz_monotone _ Hh2 (incl_refl _) Hp1).
        rewrite* map_length.
  Case "Lk_var".
    destruct (In_sig_dec l s0); try discriminate.
    inverts Hq.
    exists h'.
    split; apply* vp_postpone.
Qed.

Definition postponement h' lck' h lck e s R p q HR Hlen Hlck :=
  @sgnlz_postpone h' lck' e s R HR Hlen lck Hlck h
    (fun p q => expand_postpone (p:=p) (q:=q) _ _ _ HR Hlen Hlck) p q.

End Program.

Check correctness.
Check postponement.
